"""
fundamend contains Formate und Datenmodelle für die Energiewirtschaft in Deutschland.
pip install xml-fundamend
"""

from .models import Anwendungshandbuch, MessageImplementationGuide
from .reader import AhbReader, MigReader

__all__ = ["MigReader", "MessageImplementationGuide", "AhbReader", "Anwendungshandbuch"]



"""
the MigReader class in this module parses MIG XMLs and binds them to the data model
"""

import xml.etree.ElementTree as ET
from datetime import date, datetime
from pathlib import Path

from fundamend.models.messageimplementationguide import (
    Code,
    DataElement,
    DataElementGroup,
    MessageImplementationGuide,
    MigStatus,
    Segment,
    SegmentGroup,
)
from fundamend.reader.element_distinction import (
    _is_code,
    _is_data_element,
    _is_data_element_group,
    _is_segment,
    _is_segment_group,
)


def _to_code(element: ET.Element) -> Code:
    assert _is_code(element)
    return Code(name=element.attrib["Name"], description=element.attrib["Description"] or None, value=element.text)


def _to_data_element(element: ET.Element) -> DataElement:
    assert _is_data_element(element)
    codes = []
    for child in element:
        if _is_code(child):
            codes.append(_to_code(child))
        else:
            raise ValueError(f"unexpected element: {child.tag}")
    return DataElement(
        id=element.tag,
        name=element.attrib["Name"],
        description=element.attrib["Description"] or None,
        status_std=MigStatus(element.attrib["Status_Std"]),
        status_specification=MigStatus(element.attrib["Status_Specification"]),
        format_std=element.attrib["Format_Std"],
        format_specification=element.attrib["Format_Specification"],
        codes=codes,
    )


def _to_data_element_group(element: ET.Element) -> DataElementGroup:
    assert _is_data_element_group(element)
    data_elements = []
    for child in element:
        if _is_data_element(child):
            data_elements.append(_to_data_element(child))
        else:
            raise ValueError(f"unexpected element: {child.tag}")
    return DataElementGroup(
        id=element.tag,
        name=element.attrib["Name"],
        description=element.attrib["Description"] or None,
        status_std=MigStatus(element.attrib["Status_Std"]),
        status_specification=MigStatus(element.attrib["Status_Specification"]),
        data_elements=data_elements,
    )


def _to_segment(element: ET.Element) -> Segment:
    assert _is_segment(element)
    data_elements: list[DataElement | DataElementGroup] = []
    for child in element:
        if _is_data_element_group(child):
            data_elements.append(_to_data_element_group(child))
        elif _is_data_element(child):
            data_elements.append(_to_data_element(child))
        else:
            raise ValueError(f"unexpected element: {child.tag}")
    return Segment(
        id=element.tag.lstrip("S_"),
        name=element.attrib["Name"],
        description=element.attrib["Description"] or None,
        counter=element.attrib["Counter"],
        level=int(element.attrib["Level"]),
        max_rep_std=int(element.attrib["MaxRep_Std"]),
        max_rep_specification=int(element.attrib["MaxRep_Specification"]),
        status_std=MigStatus(element.attrib["Status_Std"]),
        status_specification=MigStatus(element.attrib["Status_Specification"]),
        example=element.attrib["Example"] or None,
        number=element.attrib["Number"],
        data_elements=data_elements,
    )


def _to_segment_group(element: ET.Element) -> SegmentGroup:
    assert _is_segment_group(element)
    segments_and_groups: list[SegmentGroup | Segment] = []
    for child in element:
        if _is_segment_group(child):
            segments_and_groups.append(_to_segment_group(child))
        elif _is_segment(child):
            segments_and_groups.append(_to_segment(child))
        else:
            raise ValueError(f"unexpected element: {child.tag}")
    return SegmentGroup(
        id=element.tag.lstrip("G_SG"),
        name=element.attrib["Name"],
        status_std=MigStatus(element.attrib["Status_Std"]),
        status_specification=MigStatus(element.attrib["Status_Specification"]),
        counter=element.attrib["Counter"],
        level=int(element.attrib["Level"]),
        max_rep_std=int(element.attrib["MaxRep_Std"]),
        max_rep_specification=int(element.attrib["MaxRep_Specification"]),
        segments=[s for s in segments_and_groups if isinstance(s, Segment)],
        segment_groups=[sg for sg in segments_and_groups if isinstance(sg, SegmentGroup)],
    )


class MigReader:
    """
    Accesses information from an XML based message implementation guide
    """

    def __init__(self, xml_path: Path):
        """
        initialize by providing the path to the XML file
        """
        self._xml_path = xml_path
        self._element_tree = ET.parse(self._xml_path)

    def get_publishing_date(self) -> date:
        """
        returns the publishing date of the message implementation guide
        """
        raw_value = self._element_tree.getroot().attrib["Veroeffentlichungsdatum"]  # e.g. '24.10.2023'
        result = datetime.strptime(raw_value, "%d.%m.%Y").date()
        return result

    def get_author(self) -> str:
        """
        returns the author of the message implementation guide
        """
        return self._element_tree.getroot().attrib["Author"]

    def get_version(self) -> str:
        """
        returns the version of the message implementation guide
        """
        return self._element_tree.getroot().attrib["Versionsnummer"]

    def get_format(self) -> str:
        """returns the format of the message implementation guide, e.g. 'UTILTS'"""
        root_tag: str = self._element_tree.getroot().tag
        return root_tag.lstrip("M_")  # converts 'M_UTILTS' to 'UTILTS'

    def _iter_segments_and_segment_groups(self, element: ET.Element) -> list[SegmentGroup | Segment]:
        """recursive function that builds a list of all segments and segment groups"""
        result: list[Segment | SegmentGroup] = []
        if _is_segment_group(element):
            result.append(_to_segment_group(element))
        elif _is_segment(element):
            result.append(_to_segment(element))
        return result

    def read(self) -> MessageImplementationGuide:
        """
        read the entire file and convert it to a MessageImplementationGuid instance
        """
        segments_and_groups = []
        for index, element in enumerate(self._element_tree.getroot()):
            if index == 0:
                continue
            segments_and_groups.extend(self._iter_segments_and_segment_groups(element))

        result = MessageImplementationGuide(
            veroeffentlichungsdatum=self.get_publishing_date(),
            autor=self.get_author(),
            versionsnummer=self.get_version(),
            format=self.get_format(),
            segments=[s for s in segments_and_groups if isinstance(s, Segment)],
            segment_groups=[s for s in segments_and_groups if isinstance(s, SegmentGroup)],
        )
        return result



import json
import pathlib

try:
    from pydantic import TypeAdapter
except ImportError as import_error:
    import_error.msg += " Did you install pydantic or fundamend[pydantic]?"
    raise
from fundamend import Anwendungshandbuch, MessageImplementationGuide

for fundamend_type in [MessageImplementationGuide, Anwendungshandbuch]:
    this_directory = pathlib.Path(__file__).parent.absolute()
    file_name = fundamend_type.__name__ + ".schema.json"  # pylint:disable=invalid-name
    file_path = this_directory / file_name
    schema = TypeAdapter(fundamend_type).json_schema()
    with open(file_path, "w", encoding="utf-8") as json_schema_file:
        json.dump(schema, json_schema_file, ensure_ascii=False, sort_keys=True, indent=4)



"""classes that represent MIGs"""

from datetime import date
from enum import StrEnum

from ._dataclass_wrapper import dataclass

# I didn't invent the data model ;)
# pylint:disable=too-many-instance-attributes


class MigStatus(StrEnum):
    """
    status of a MIG element
    """

    M = "M"
    C = "C"
    R = "R"
    N = "N"
    D = "D"


@dataclass(kw_only=True, eq=True)
class Code:
    """
    a single code element inside a MIG Dataelement
    """

    # Example:
    # <Code Name="Netznutzungszeiten-Nachricht" Description="">UTILTS</Code>
    name: str  # e.g. 'Netznutzungszeiten-Nachricht'
    description: str | None = None  # e.g. ''
    value: str | None  # e.g. 'UTILTS'


@dataclass(kw_only=True, eq=True)
class DataElement:
    """
    A single data element inside a MIG Segment.
    This models both the 'Datenelement' and the 'Gruppendatenelement'
    """

    # pylint:disable=line-too-long
    # Example:
    # <D_0065 Name="Nachrichtentyp-Kennung" Description="" Status_Std="M" Status_Specification="M" Format_Std="an..6" Format_Specification="an..6">
    #   <Code Name="Netznutzungszeiten-Nachricht" Description="">UTILTS</Code>
    # </D_0065>
    id: str  # e.g. 'D_0065'
    name: str  # e.g. 'Nachrichtentyp-Kennung'
    description: str | None = None  # e.g. ''
    status_std: MigStatus
    status_specification: MigStatus
    format_std: str  #: e.g. 'an..6'
    format_specification: str  #: e.g. 'an..6'
    codes: list[Code]


@dataclass(eq=True, kw_only=True)
class DataElementGroup:
    """
    a group of data elements, German 'Datenelementgruppe'.
    """

    # "Die Datenelementgruppe C0829 enthält mehrere Gruppendatenelemente. Diese Datenelementgruppe enthält das
    # Gruppendatenelement DE3039, hier wird die MP-ID angegeben, sowie das DE3055, welches die code-vergebende Stelle
    # definiert. Diese Datenelementgruppe enthält des Weiteren das Gruppendatenelement DE113110, welches nur in der MIG
    # und nicht im AHB aufgeführt wird, um den Aufbau der EDIFACT Nachricht korrekt umsetzen zu können."
    # Quelle: Allgemeine Festlegungen Kapitel 6.1, Seite 52

    # Example:
    # pylint:disable=line-too-long
    # <C_C082 Name="Identifikation des Beteiligten" Description="" Status_Std="C" Status_Specification="R">
    #   <D_3039 Name="MP-ID" Description="" Status_Std="M" Status_Specification="M" Format_Std="an..35" Format_Specification="an..35"/>
    #   <D_1131 Name="Codeliste, Code" Description="" Status_Std="C" Status_Specification="N" Format_Std="an..17" Format_Specification="an..17"/>
    #   <D_3055 Name="Verantwortliche Stelle für die Codepflege, Code" Description="" Status_Std="C" Status_Specification="R" Format_Std="an..3" Format_Specification="an..3">
    #     <Code Name="GS1" Description="">9</Code>
    #     <Code Name="DE, BDEW (Bundesverband der Energie- und Wasserwirtschaft e.V.)" Description="">293</Code>
    #   </D_3055>
    # </C_C082>

    id: str  # e.g. 'C_C082'
    name: str  # e.g. 'Identifikation des Beteiligten'
    description: str | None = None  # e.g. ''
    status_std: MigStatus
    status_specification: MigStatus
    data_elements: list[DataElement]


@dataclass(frozen=True, eq=True, order=True, unsafe_hash=True, kw_only=True)
class Segment:
    """
    a segment inside a MIG
    """

    # Example:
    # pylint:disable=line-too-long
    # <S_NAD Name="MP-ID Absender" Description="DE3039: Zur Identifikation der Marktpartner wird die MP-ID angegeben." Counter="0100" Level="1" Number="00004" MaxRep_Std="1" MaxRep_Specification="1" Status_Std="M" Status_Specification="M" Example="NAD+MS+9900259000002::293&apos;">
    #     <D_3035 Name="Beteiligter, Qualifier" Description="" Status_Std="M" Status_Specification="M" Format_Std="an..3" Format_Specification="an..3">
    #        <Code Name="Dokumenten-/Nachrichtenaussteller bzw. -absender" Description="">MS</Code>
    #     </D_3035>
    #     <C_C082 Name="Identifikation des Beteiligten" Description="" Status_Std="C" Status_Specification="R">
    #       <D_3039 Name="MP-ID" Description="" Status_Std="M" Status_Specification="M" Format_Std="an..35" Format_Specification="an..35"/>
    #       <D_1131 Name="Codeliste, Code" Description="" Status_Std="C" Status_Specification="N" Format_Std="an..17" Format_Specification="an..17"/>
    #       <D_3055 Name="Verantwortliche Stelle für die Codepflege, Code" Description="" Status_Std="C" Status_Specification="R" Format_Std="an..3" Format_Specification="an..3">
    #         <Code Name="GS1" Description="">9</Code>
    #         <Code Name="DE, BDEW (Bundesverband der Energie- und Wasserwirtschaft e.V.)" Description="">293</Code>
    #      </D_3055>
    #   </C_C082>
    # </S_NAD>
    id: str  #: e.g. 'NAD'
    name: str  #: e.g. 'MP-ID Absender'
    description: str | None = None  # e.g. 'DE3039: Zur Identifikation der Marktpartner wird die MP-ID angegeben.'
    counter: str  #: e.g. '0100'
    level: int  #: e.g. 1
    number: str  #: e.g. '00004'
    max_rep_std: int  #: e.g. 1
    max_rep_specification: int  #: e.g. 1
    status_std: MigStatus
    status_specification: MigStatus
    example: str | None  #: e.g. "NAD+MS+9900259000002::293'"
    data_elements: list[DataElement | DataElementGroup]


@dataclass(kw_only=True, eq=True)
class SegmentGroup:
    """
    a "Segtmentgruppe"
    """

    # pylint:disable=line-too-long
    # Example:
    # <G_SG2 Name="MP-ID Empfänger" Counter="0090" Level="1" MaxRep_Std="99" MaxRep_Specification="1" Status_Std="C" Status_Specification="R">
    #   <S_NAD Name="MP-ID Empfänger" Description="DE3039: Zur Identifikation der Marktpartner wird die MP-ID angegeben." Counter="0100" Level="1" Number="00007" MaxRep_Std="1" MaxRep_Specification="1" Status_Std="M" Status_Specification="M" Example="NAD+MR+9900259000002::293&apos;">
    #     <D_3035 Name="Beteiligter, Qualifier" Description="" Status_Std="M" Status_Specification="M" Format_Std="an..3" Format_Specification="an..3">
    #        <Code Name="Nachrichtenempfänger" Description="">MR</Code>
    #     </D_3035>
    #     <C_C082 Name="Identifikation des Beteiligten" Description="" Status_Std="C" Status_Specification="R">
    #       <D_3039 Name="MP-ID" Description="" Status_Std="M" Status_Specification="M" Format_Std="an..35" Format_Specification="an..35"/>
    #       <D_1131 Name="Codeliste, Code" Description="" Status_Std="C" Status_Specification="N" Format_Std="an..17" Format_Specification="an..17"/>
    #       <D_3055 Name="Verantwortliche Stelle für die Codepflege, Code" Description="" Status_Std="C" Status_Specification="R" Format_Std="an..3" Format_Specification="an..3">
    #         <Code Name="GS1" Description="">9</Code>
    #         <Code Name="DE, BDEW (Bundesverband der Energie- und Wasserwirtschaft e.V.)" Description="">293</Code>
    #       </D_3055>
    #     </C_C082>
    #   </S_NAD>
    # </G_SG2>
    id: str  # e.g. 'SG2'
    name: str  # e.g. 'MP-ID Empfänger'
    counter: str  # e.g. '0090'
    level: int  # e.g. 1
    max_rep_std: int  # e.g. 99
    max_rep_specification: int  # e.g. 1
    status_std: MigStatus
    status_specification: MigStatus
    segments: list[Segment]
    segment_groups: list["SegmentGroup"]


@dataclass(kw_only=True, eq=True)
class MessageImplementationGuide:
    """
    message implementation guide (MIG)
    """

    veroeffentlichungsdatum: date
    """publishing date"""

    autor: str
    """author, most likely 'BDEW'"""

    versionsnummer: str
    """e.g. '1.1c'"""

    format: str  #: e.g. 'UTILTS'

    segments: list[Segment]
    segment_groups: list[SegmentGroup]



"""
the AhbReader class in this module parses AHB XMLs and binds them to the data model
"""

import re
import xml.etree.ElementTree as ET
from datetime import date, datetime
from pathlib import Path

from fundamend.models.anwendungshandbuch import (
    Anwendungsfall,
    Anwendungshandbuch,
    Bedingung,
    Code,
    DataElement,
    DataElementGroup,
    Paket,
    Segment,
    SegmentGroup,
    UbBedingung,
)
from fundamend.reader.element_distinction import (
    _is_anwendungsfall,
    _is_code,
    _is_data_element,
    _is_data_element_group,
    _is_format,
    _is_segment,
    _is_segment_group,
)

# pylint:disable=duplicate-code
# yes, it's very similar to the MigReader


def _to_code(element: ET.Element) -> Code:
    assert _is_code(element)
    return Code(
        name=element.attrib["Name"],
        description=element.attrib["Description"] or None,
        value=element.text,
        ahb_status=element.attrib["AHB_Status"],
    )


def _to_bedingung(element: ET.Element) -> Bedingung:
    return Bedingung(
        nummer=element.attrib["Nummer"].lstrip("[").rstrip("]"),
        text=element.text.strip(),  # type:ignore[union-attr]
    )


def _to_ub_bedingung(element: ET.Element) -> UbBedingung:
    return UbBedingung(
        nummer=element.attrib["Nummer"].lstrip("[").rstrip("]"),
        text=element.text.strip(),  # type:ignore[union-attr]
    )


def _to_paket(element: ET.Element) -> Paket:
    return Paket(
        nummer=element.attrib["Nummer"].lstrip("[").rstrip("]"),
        text=element.text.strip(),  # type:ignore[union-attr]
    )


_pruefi_pattern = re.compile(r"\d{5}")


def _is_valid_pruefidentifikator(pruefidentifikator: str) -> bool:
    """
    returns true if the passed object looks like a pruefinentifikator
    """
    if pruefidentifikator is None:
        return False
    if not isinstance(pruefidentifikator, str):
        return False  # this prevents users from accidentally passing integers
    return _pruefi_pattern.match(pruefidentifikator) is not None


def _to_data_element(element: ET.Element) -> DataElement:
    assert _is_data_element(element)
    codes = []
    for child in element:
        if _is_code(child):
            codes.append(_to_code(child))
        else:
            raise ValueError(f"unexpected element: {child.tag}")
    return DataElement(
        id=element.tag,
        name=element.attrib["Name"],
        codes=codes,
    )


def _to_data_element_group(element: ET.Element) -> DataElementGroup:
    assert _is_data_element_group(element)
    data_elements = []
    for child in element:
        if _is_data_element(child):
            data_elements.append(_to_data_element(child))
        else:
            raise ValueError(f"unexpected element: {child.tag}")
    return DataElementGroup(
        id=element.tag,
        name=element.attrib["Name"],
        data_elements=data_elements,
    )


def _to_segment(element: ET.Element) -> Segment:
    assert _is_segment(element)
    data_elements: list[DataElement | DataElementGroup] = []
    for child in element:
        if _is_data_element_group(child):
            data_elements.append(_to_data_element_group(child))
        elif _is_data_element(child):
            data_elements.append(_to_data_element(child))
        else:
            raise ValueError(f"unexpected element: {child.tag}")
    ahb_status: str | None = None
    if "AHB_Status" in element.attrib and element.attrib["AHB_Status"].strip():
        ahb_status = element.attrib["AHB_Status"]
    return Segment(
        id=element.tag.lstrip("S_"),
        name=element.attrib["Name"],
        number=element.attrib["Number"],
        ahb_status=ahb_status,
        data_elements=data_elements,
    )


def _to_segment_group(element: ET.Element) -> SegmentGroup:
    assert _is_segment_group(element)
    segments_and_groups: list[SegmentGroup | Segment] = []
    for child in element:
        if _is_segment_group(child):
            segments_and_groups.append(_to_segment_group(child))
        elif _is_segment(child):
            segments_and_groups.append(_to_segment(child))
        else:
            raise ValueError(f"unexpected element: {child.tag}")
    return SegmentGroup(
        id=element.tag.lstrip("G_SG"),
        name=element.attrib["Name"],
        ahb_status=element.attrib["AHB_Status"],
        segments=[s for s in segments_and_groups if isinstance(s, Segment)],
        segment_groups=[sg for sg in segments_and_groups if isinstance(sg, SegmentGroup)],
    )


class AhbReader:
    """
    Accesses information from an XML based Anwendungshandbuch
    """

    def __init__(self, xml_path: Path):
        """
        initialize by providing the path to the XML file
        """
        self._xml_path = xml_path
        self._element_tree = ET.parse(self._xml_path)

    def get_publishing_date(self) -> date:
        """
        returns the publishing date of the message implementation guide
        """
        raw_value = self._element_tree.getroot().attrib["Veroeffentlichungsdatum"]  # e.g. '02.04.2024'
        result = datetime.strptime(raw_value, "%d.%m.%Y").date()
        return result

    def get_author(self) -> str:
        """
        returns the author of the AHB
        """
        return self._element_tree.getroot().attrib["Author"]

    def get_version(self) -> str:
        """
        returns the version of the AHB
        """
        return self._element_tree.getroot().attrib["Versionsnummer"]

    def get_bedingungen(self) -> list[Bedingung]:
        """returns the plain bedingungen"""
        return [_to_bedingung(x) for x in self._element_tree.getroot().find("Bedingungen")]  # type:ignore[union-attr]

    def get_ub_bedingungen(self) -> list[UbBedingung]:
        """returns the UB Bedingungen"""
        return [
            _to_ub_bedingung(x) for x in self._element_tree.getroot().find("UB_Bedingungen")  # type:ignore[union-attr]
        ]

    def get_pakete(self) -> list[Paket]:
        """returns the package definitions"""
        return [_to_paket(x) for x in self._element_tree.getroot().find("Pakete")]  # type:ignore[union-attr]

    def get_anwendungsfall(self, pruefidentifikator: str) -> Anwendungsfall | None:
        """find the anwendungsfall matching the pruefidentifikator or return None"""
        if not _is_valid_pruefidentifikator(pruefidentifikator):
            raise ValueError(f"invalid pruefidentifikator: {pruefidentifikator}")
        for element in self._element_tree.getroot():
            if element.tag != "AWF":
                continue
            if element.attrib["Pruefidentifikator"] != pruefidentifikator:
                continue
            if element.tag == "AWF" and element.attrib["Pruefidentifikator"] == pruefidentifikator:
                return self._read_anwendungsfall(element)
        return None

    def get_anwendungsfaelle(self) -> list[Anwendungsfall]:
        """finds all anwendungsfaelle in the XML file"""
        result: list[Anwendungsfall] = []
        for element in self._element_tree.getroot():
            if element.tag != "AWF":
                continue
            result.append(self._read_anwendungsfall(element))
        return result

    def _iter_segments_and_segment_groups(self, element: ET.Element) -> list[SegmentGroup | Segment]:
        """recursive function that builds a list of all segments and segment groups"""
        result: list[Segment | SegmentGroup] = []
        if _is_anwendungsfall(element) or _is_format(element):
            for sub_element in element:
                result.extend(self._iter_segments_and_segment_groups(sub_element))
        if _is_segment_group(element):
            result.append(_to_segment_group(element))
        elif _is_segment(element):
            result.append(_to_segment(element))
        return result

    def _read_anwendungsfall(self, original_element: ET.Element) -> Anwendungsfall:
        segments_and_groups = []
        for element in self._element_tree.getroot():
            segments_and_groups.extend(self._iter_segments_and_segment_groups(element))
        return Anwendungsfall(
            pruefidentifikator=original_element.attrib["Pruefidentifikator"],
            beschreibung=original_element.attrib["Beschreibung"],
            kommunikation_von=original_element.attrib["Kommunikation_von"],
            format=original_element.tag.lstrip("M_"),
            segments=[s for s in segments_and_groups if isinstance(s, Segment)],
            segment_groups=[s for s in segments_and_groups if isinstance(s, SegmentGroup)],
        )

    def read(self) -> Anwendungshandbuch:
        """
        read the entire file and convert it to a MessageImplementationGuid instance
        """

        result = Anwendungshandbuch(
            veroeffentlichungsdatum=self.get_publishing_date(),
            autor=self.get_author(),
            versionsnummer=self.get_version(),
            bedingungen=self.get_bedingungen(),
            ub_bedingungen=self.get_ub_bedingungen(),
            pakete=self.get_pakete(),
            anwendungsfaelle=self.get_anwendungsfaelle(),
        )
        return result



"""
wrapper around either pydantic.dataclass or stdlib dataclass
"""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from dataclasses import dataclass
else:
    try:
        from pydantic.dataclasses import dataclass  # pylint:disable=unused-import
    except ImportError:
        from dataclasses import dataclass
__all__ = ["dataclass"]



""""model classes for Anwendungshandbücher (AHB)"""

# pylint:disable=duplicate-code
# the structures are similar, still we decided against inheritance, so there's naturally a little bit of duplication

from datetime import date

from ._dataclass_wrapper import dataclass


@dataclass(kw_only=True, eq=True, frozen=True)
class Code:
    """
    a single code element inside an AHB DataElement
    """

    # Example:
    # <Code Name="Netznutzungszeiten-Nachricht" Description="" AHB_Status="X">UTILTS</Code>
    # properties are similar to MIG data model, still we like to keep them separate instead of inheriting
    name: str  # e.g. 'Netznutzungszeiten-Nachricht'
    description: str | None = None  # e.g. ''
    value: str | None  # e.g. 'UTILTS'
    ahb_status: str  #: e.g. 'X' # new for AHB


@dataclass(kw_only=True, eq=True, frozen=True)
class DataElement:
    """
    A single data element inside a AHB Segment.
    This models both the 'Datenelement' and the 'Gruppendatenelement'
    """

    # Example:
    # <D_0065 Name="Nachrichtentyp-Kennung">
    #   <Code Name="Netznutzungszeiten-Nachricht" Description="" AHB_Status="X">UTILTS</Code>
    # </D_0065>
    id: str  # e.g. 'D_0065'
    name: str  # e.g. 'Nachrichtentyp-Kennung'
    codes: list[Code]


@dataclass(eq=True, kw_only=True, frozen=True)
class DataElementGroup:
    """
    a group of data elements, German 'Datenelementgruppe' inside the AHB
    """

    # "Die Datenelementgruppe C0829 enthält mehrere Gruppendatenelemente. Diese Datenelementgruppe enthält das
    # Gruppendatenelement DE3039, hier wird die MP-ID angegeben, sowie das DE3055, welches die code-vergebende Stelle
    # definiert. Diese Datenelementgruppe enthält des Weiteren das Gruppendatenelement DE113110, welches nur in der MIG
    # und nicht im AHB aufgeführt wird, um den Aufbau der EDIFACT Nachricht korrekt umsetzen zu können."
    # Quelle: Allgemeine Festlegungen Kapitel 6.1, Seite 52

    # Example:
    # <C_C002 Name="Dokumenten-/Nachrichtenname">
    #    <D_1001 Name="Dokumentenname, Code">
    #      <Code Name="Berechnungsformel" Description="" AHB_Status="X">Z36</Code>
    #   </D_1001>
    # </C_C002>

    id: str  # e.g. 'C_C082'
    name: str  # e.g. 'Dokumenten-/Nachrichtenname'
    data_elements: list[DataElement]


@dataclass(frozen=True, eq=True, unsafe_hash=True, kw_only=True)
class Segment:
    """
    a segment inside an AHB
    """

    # Example:
    # pylint:disable=line-too-long
    #  <S_BGM Name="Beginn der Nachricht" Number="00002" AHB_Status="Muss">
    #    <C_C002 Name="Dokumenten-/Nachrichtenname">
    #       <D_1001 Name="Dokumentenname, Code">
    #         <Code Name="Berechnungsformel" Description="" AHB_Status="X">Z36</Code>
    #       </D_1001>
    #    </C_C002>
    #    <C_C106 Name="Dokumenten-/Nachrichten-Identifikation">
    #       <D_1004 Name="Dokumentennummer" AHB_Status="X"/>
    #    </C_C106>
    # </S_BGM>
    id: str  #: e.g. 'BGM'
    name: str  #: e.g. 'Beginn der Nachricht'
    number: str  #: e.g. '00002'
    ahb_status: str | None  #: e.g. 'Muss'
    data_elements: list[DataElement | DataElementGroup]


@dataclass(kw_only=True, eq=True, frozen=True)
class SegmentGroup:
    """
    a "Segmentgruppe" inside an AHB
    """

    # Example:
    # <G_SG6 Name="Prüfidentifikator" AHB_Status="Muss">
    #   <S_RFF Name="Prüfidentifikator" Number="00012" AHB_Status="Muss">
    #     <C_C506 Name="Referenz">
    #       <D_1153 Name="Referenz, Qualifier">
    #         <Code Name="Prüfidentifikator" Description="" AHB_Status="X">Z13</Code>
    #       </D_1153>
    #       <D_1154 Name="Referenz, Identifikation">
    #         <Code Name="Berechnungsformel" Description="" AHB_Status="X">25001</Code>
    #       </D_1154>
    #     </C_C506>
    #   </S_RFF>
    #  </G_SG6>
    id: str  #: e.g. 'SG6'
    name: str  #: e.g. 'Prüfidentifikator'
    ahb_status: str | None  #: e.g. 'Muss'
    segments: list[Segment]
    segment_groups: list["SegmentGroup"]


@dataclass(kw_only=True, eq=True, frozen=True)
class Anwendungsfall:
    """One Anwendungsfall "AWF" corresponds to one Prüfidentifikator or Type of Message"""

    # Example:
    # <AWF Pruefidentifikator="25001" Beschreibung="Berechnungsformel" Kommunikation_von="NB an MSB / LF">
    #   <M_UTILTS>
    #     <S_UNH Name="Nachrichten-Kopfsegment" Number="00001" AHB_Status="Muss">
    #    ...
    #     </S_UNT>
    #   </M_UTILTS>
    # </AWF>
    pruefidentifikator: str  #: e.g. '25001'
    beschreibung: str  #: e.g. 'Berechnungsformel'
    kommunikation_von: str  #: e.g. 'NB an MSB / LF'
    format: str  #: e.g. 'UTILTS'
    segments: list[Segment]
    segment_groups: list[SegmentGroup]


@dataclass(kw_only=True, eq=True, frozen=True)
class Bedingung:
    """Ein ConditionKeyConditionText Mapping"""

    nummer: str  #: e.g. '1'
    text: str  #: e.g. 'Nur MP-ID aus Sparte Strom'


@dataclass(kw_only=True, eq=True, frozen=True)
class UbBedingung:
    """Eine UB-Bedingung"""

    # Example:
    # <UB_Bedingung Nummer="[UB1]">([931] ∧ [932] [490]) ⊻ ([931] ∧ [933] [491])</UB_Bedingung>
    nummer: str  #: e.g. 'UB1'
    text: str  #: e.g. '([931] ∧ [932] [490]) ⊻ ([931] ∧ [933] [491])'


@dataclass(kw_only=True, eq=True, frozen=True)
class Paket:
    """Ein Bedingungspaket/PackageKeyConditionText Mapping"""

    # Example:
    # <Paket Nummer="[1P]">--</Paket>
    nummer: str  #: e.g. '1P'
    text: str  #: e.g. '--'


@dataclass(kw_only=True, eq=True, frozen=True)
class Anwendungshandbuch:
    """
    Ein Anwendungshandbuch bündelt verschiedene Nachrichtentypen/Anwendungsfälle im selben Format oder mit der selben
    regulatorischen Grundlage und stellt gemeinsame Pakete & Bedingungen bereit.
    """

    # Example:
    # <AHB Versionsnummer="1.1d" Veroeffentlichungsdatum="02.04.2024" Author="BDEW">
    #  <AWF Pruefidentifikator="25001" Beschreibung="Berechnungsformel" Kommunikation_von="NB an MSB / LF">
    #  </AWF>
    # </AHB>
    veroeffentlichungsdatum: date
    """publishing date"""

    autor: str
    """author, most likely 'BDEW'"""

    versionsnummer: str
    """e.g. '1.1d'"""
    anwendungsfaelle: list[Anwendungsfall]  #: die einzelnen Prüfidendifikatoren
    bedingungen: list[Bedingung]
    ub_bedingungen: list[UbBedingung]
    pakete: list[Paket]
